/*
 * The main object for ruling the history blocking methods and local variables.
 */
var historyblock = {

  //
  // Local variables for this historyblock object.
  //
  downloadManager: null,
  hostRegexp: null,
  goodHostRegexp: null,
  slashesRegexp: null,
  whiteSpace: null,
  localprefs: null,
  prefs: null,
  domainstring: null,
  historyManager: null,
  observerService: null,
  hexcase: null,
  b64pad: null,
  chrsz: null,
  navHistoryService: null,
  pbs: null,
  cookieManager: null,
  blockCookies: null,
  STATE_START: Components.interfaces.nsIWebProgressListener.STATE_START,
  STATE_STOP: Components.interfaces.nsIWebProgressListener.STATE_STOP,


  //
  // Main Methods
  //
  
  /*
   * The init function sets up all the local variables and listeners.
   */
  init: function() 
  {
	historyblock.whiteSpace = /\s/;
	historyblock.goodHostRegexp = /\,/;
    // Special thanks to Viktar Karpach for this bit of regular expressions.
	historyblock.slashesRegexp = /(^([^:]*:\/\/)?(www\.)?)/;
	historyblock.hostRegexp = /^([^:]*:\/\/)?[^\/]*/;
	
	historyblock.downloadManager = 
	        Components.classes["@mozilla.org/download-manager;1"]
	                  .createInstance(Components.interfaces.nsIDownloadManager);
    historyblock.hexcase = 0;  /* hex output format. 0 - lowercase; 1 - uppercase        */
    historyblock.b64pad  = ""; /* base-64 pad character. "=" for strict RFC compliance   */
    historyblock.chrsz   = 8;  /* bits per input character. 8 - ASCII; 16 - Unicode      */

    historyblock.navHistoryService =
        Components.classes["@mozilla.org/browser/nav-history-service;1"]
                  .getService(Components.interfaces.nsINavHistoryService);
                  
    historyblock.navHistoryService.addObserver(historyblock, false);
    
    historyblock.prefs = 
        Components.classes["@mozilla.org/preferences-service;1"]
                  .getService(Components.interfaces.nsIPrefBranch);

    historyblock.domainstring = historyblock.prefs.getCharPref("extensions.historyblock.stringpref");

    this.localprefs = 
        Components.classes["@mozilla.org/preferences-service;1"]
                  .getService(Components.interfaces.nsIPrefService)
                  .getBranch("extensions.historyblock.");   

    this.localprefs.QueryInterface(Components.interfaces.nsIPrefBranch2);   
    this.localprefs.addObserver("", historyblock, false);
    
    historyblock.historyManager = 
        Components.classes["@mozilla.org/browser/nav-history-service;1"]
                  .getService(Components.interfaces.nsIBrowserHistory);

    historyblock.observerService = 
        Components.classes["@mozilla.org/observer-service;1"]
                  .getService(Components.interfaces.nsIObserverService);  
    
    historyblock.observerService.addObserver(historyblock,"dl-done",false);

    window.addEventListener("unload", function() {
      observerService.removeObserver(historyblock, "dl-done");
    }, false);
    
    gBrowser.tabContainer.addEventListener("TabClose",historyblock.tabClosed, true); 

    // FF3.5 specific code
    historyblock.pbs = Components.classes["@mozilla.org/privatebrowsing;1"]  
    	                         .getService(Components.interfaces.nsIPrivateBrowsingService);
    
    historyblock.cookieManager = Components.classes["@mozilla.org/cookiemanager;1"]
    									   .getService(Components.interfaces.nsICookieManager2);
    
    gBrowser.addTabsProgressListener(historyblock);
  },
  
  /**
   * Apparently required by the mozilla addons folks...
   */
  onLocationChange: function(aBrowser, webProgress, request, location){},
  onProgressChange: function(aBrowser, webProgress, request, curSelfProgress, maxSelfProgress, curTotalProgress, maxTotalProgress){},
  onSecurityChange: function(aBrowser, aWebProgress, aRequest, aState){},
  onStatusChange: function(aBrowser, aWebProgress, aRequest, aStatus, aMessage){},
  onRefreshAttempted: function(aBrowser, webProgress, aRefreshURI, aMillis, aSameURI){return true;},
  onLinkIconAvailable: function(aBrowser){},
  
  /**
   * Gets fired every time the state changes on ANY page on ANY tab in THIS windows.
   */
  onStateChange: function(aBrowser, aWebProgress, aRequest, aStateFlag, aStatus)
  {
	with(historyblock)
	{
	  if(aStateFlag & STATE_STOP)
	  {
		// Only gets into this if statement when a page is completely done rendering.
		// Fires on HTML, XML, GIFs, etc.
	    if(aStatus == 0)
	    {
	  	  var domains = domainstring.split(",");
		  var hashedTargetUrl = aBrowser.currentURI.spec;
		
		  hashedTargetUrl = hashedTargetUrl.match(hostRegexp);
		  
		  if(hashedTargetUrl == null) return; // should never happen
		  var hostname = hashedTargetUrl[0].replace(slashesRegexp,"");
		
		  removeHistory(hostname);
	    }
	  }
	}
  },
   
  /*
   * Gets called whenever a tab has closed.
   */
  tabClosed: function(event)
  {
    with(historyblock)
    {
      var domains = domainstring.split(",");
      var hashedTargetUrl = event.target.linkedBrowser.currentURI.spec;

      hashedTargetUrl = hashedTargetUrl.match(hostRegexp);
      if(hashedTargetUrl == null) return; // should never happen
      var hostname = hashedTargetUrl[0].replace(slashesRegexp,"");
      
      // Split on the '.'s
      var subdomains = hostname.split(".");
      if(subdomains.length > 2)
      {
    	  // Then we have some subdomain-ing going on.
    	  if(prefs.getBoolPref("extensions.historyblock.blocksitewide"))
    	  {
    		  subdomains = subdomains[subdomains.length - 2] + "." + subdomains[subdomains.length - 1];
    	  }
      }
      else
      {
    	  // Break it
    	  subdomains = null;
      }
      
      // Hash that hostname
      hashedTargetUrl = hex_sha1(hostname);
      // Walk our blacklisted domains
      for(var blIndex=0; blIndex<domains.length; blIndex++)
      {
        // If the hashed URL of the closed tab contains a hashed blacklisted hostname...
        if(hashedTargetUrl == domains[blIndex].replace(whiteSpace,"")) 
        {
          // Before we wipe all the data, let's collect the cookies
          var blacklistCookies = getBlacklistedCookies(hostname);
          // The new hotness that is FF3.5 api
          pbs.removeDataFromDomain(hostname);
          // Okay, we've wiped everything, including the cookies... we MIGHT
          // want to keep those, however, so let's add them back.
          readdBlacklistCookies(blacklistCookies);
          return;
        }
      }
      // If we got here, there was no match, try to match the site-wide-domain if possible
      if(subdomains == null) return;      
      hashedTargetUrl = hex_sha1(subdomains);
      // Walk our blacklisted domains
      for(var blIndex=0; blIndex<domains.length; blIndex++)
      {
        // If the hashed URL of the closed tab contains a hashed blacklisted hostname...
        if(hashedTargetUrl == domains[blIndex].replace(whiteSpace,"")) 
        {
          // Before we wipe all the data, let's collect the cookies via hostname
          var blacklistCookies = getBlacklistedCookies(hostname);
          // The new hotness that is FF3.5 api
          pbs.removeDataFromDomain(hostname);
          // Okay, we've wiped everything, including the cookies... we MIGHT
          // want to keep those, however, so let's add them back.
          readdBlacklistCookies(blacklistCookies);
          return;
        }
      }
    }
  },
  
  /*
   * Gets called when the user clicks "Block This!" from the right-click menu.
   */
  onBlockThis: function()
  {
    with (historyblock)
    {
      var hostname = getBrowser().currentURI.spec;

      hostname = hostname.match(hostRegexp);
      hostname = hostname[0].replace(slashesRegexp,"");
      if(confirm('Do you really want to block "' + hostname + '"?'))
      {
        // set the temp variable that will fire the rest
        prefs.setCharPref("extensions.historyblock.newblockedurl", hostname);
        // Before we wipe all the data, let's collect the cookies
        var blacklistCookies = getBlacklistedCookies(hostname);
        removeHistory(hostname);
        // The new hotness that is FF3.5 api
        pbs.removeDataFromDomain(hostname);
        // Okay, we've wiped everything, including the cookies... we MIGHT
        // want to keep those, however, so let's add them back.
        readdBlacklistCookies(blacklistCookies);
        return;
      }
    }
  },
  
  removeHistory: function(hostname)
  {
	with (historyblock)
	{
      // Split on the '.'s
      var subdomains = hostname.split(".");
      if(subdomains.length > 2)
      {
		// Then we have some subdomain-ing going on.
		if(prefs.getBoolPref("extensions.historyblock.blocksitewide"))
		{
		  subdomains = subdomains[subdomains.length - 2] + "." + subdomains[subdomains.length - 1];
		}
      }
      else
      {
    	// Break it
    	subdomains = null;
      }
      
      // Hash that hostname
      hashedTargetUrl = hex_sha1(hostname);
        
      var domains = domainstring.split(",");
      // Walk our blacklisted domains
      for(var blIndex=0; blIndex<domains.length; blIndex++)
      {
        // If the hashed URL of the visited tab contains a hashed blacklisted hostname...
        if(hashedTargetUrl == domains[blIndex].replace(whiteSpace,"")) 
        {
          historyManager.removePagesFromHost(hostname,true);
          return;
        }
      }
      // If we got here, there was no match, try to match the site-wide-domain if possible
      if(subdomains == null) return;
      hashedTargetUrl = hex_sha1(subdomains);
      // Walk our blacklisted domains
      for(var blIndex=0; blIndex<domains.length; blIndex++)
      {
        // If the hashed URL of the visited tab contains a hashed blacklisted hostname...
        if(hashedTargetUrl == domains[blIndex].replace(whiteSpace,"")) 
        {
          historyManager.removePagesFromHost(subdomains,true);
          return;
        }
      }
	}
  },
  
  // These are only placeholder listeners, errors get thrown if these don't exist.
  onVisit: function(aURI, aVisitID, aTime, aSessionID, aReferringID, aTransitionType){},
  onBeginUpdateBatch: function(){return 0;},
  onClearHistory: function(){return 0;},
  onDeleteURI: function(aURI){return 0;},
  onEndUpdateBatch: function(){return 0;},
  onPageChanged: function(aURI,aWhat,aValue){return 0;},
  onPageExpired: function(aURI,aVisitTime,aWholeEntry){return 0;},
  onTitleChanged: function(aURI,aPageTitle){return 0;},
  
  /*
   * Not actually listened for explicitly, but rather gets called when the observe
   * method finds adding a hostname applicable.
   */
  onHostnameAdd: function()
  {
    with (historyblock)
    {
      // Get the new blocked host
      var newBlockedHost = prefs.getCharPref("extensions.historyblock.newblockedurl");
      
      newBlockedHost = newBlockedHost.match(hostRegexp);
      if(newBlockedHost == null) return;
      newBlockedHost = newBlockedHost[0].replace(slashesRegexp,"");
      
      // Check for bad values and return on them.
      if(newBlockedHost == "") return;
      var domains = domainstring.split(",");
      for(var i=0; i<domains.length; i++)
      {
        if(domains[i].replace(whiteSpace,"") == hex_sha1(newBlockedHost))
        {
          prefs.setCharPref("extensions.historyblock.newblockedurl", "");
          return;
        }
      }
      // Made it past the sentinel, we have a good hostname that isn't a repeat.
      // Remove it from the history to start, just to have a blank starting point.
      // Before we wipe all the data, let's collect the cookies
      var blacklistCookies = getBlacklistedCookies(newBlockedHost);
      // The new hotness that is FF3.5 api
      pbs.removeDataFromDomain(newBlockedHost);
      // Okay, we've wiped everything, including the cookies... we MIGHT
      // want to keep those, however, so let's add them back.
      readdBlacklistCookies(blacklistCookies);
      // Hash the newBlockedHost for privacy's sake.
      newBlockedHost = hex_sha1(newBlockedHost);
      // Add it to the local prefs
      domainstring += (", "+newBlockedHost);
      domainstring = domainstring.replace(",,",",");
      domainstring = domainstring.replace(", ,",",");
      if(domainstring.length >= 2)
      {
	      if(domainstring.substring(0,2) == ' ,') domainstring = domainstring.substring(1,domainstring.length);
	      if(domainstring.substring(0,1) == ',') domainstring = domainstring.substring(1,domainstring.length);
	      if(domainstring.substring(domainstring.length-2,domainstring.length) == ', ') domainstring = domainstring.substring(0,domainstring.length-2);
      }
      // Set the new prefs
      prefs.setCharPref("extensions.historyblock.stringpref", domainstring);
      // Set the newblockedurl to null
      prefs.setCharPref("extensions.historyblock.newblockedurl", "");
    }
  },
  
  /*
   * This is called whenever a hostname is entered with the remove checkbox true.
   */
  onHostnameRemove: function()
  {
    with(historyblock)
    {
      // Get the new blocked host
      var unBlockedHost = prefs.getCharPref("extensions.historyblock.unblockedurl");
      
      unBlockedHost = unBlockedHost.match(hostRegexp);
      if(unBlockedHost == null) return;
      unBlockedHost = unBlockedHost[0].replace(slashesRegexp,"");
      
      // Check for bad values and return on them.
      if(unBlockedHost == "") return;
      // We're unblacklisting
      var domains = domainstring.split(",");
      for(var i=0; i<domains.length; i++)
      {
        if(domains[i].replace(whiteSpace,"") == hex_sha1(unBlockedHost))
        {
          var removed = domains[i].replace(whiteSpace,"");
          domainstring = domainstring.replace(removed,"");
          domainstring = domainstring.replace(",,",",");
          domainstring = domainstring.replace(", ,",",");
          if(domainstring.substring(0,2) == ' ,') domainstring = domainstring.substring(1,domainstring.length);
          if(domainstring.substring(0,1) == ',') domainstring = domainstring.substring(1,domainstring.length);
          if(domainstring.substring(domainstring.length-2,domainstring.length) == ', ') domainstring = domainstring.substring(0,domainstring.length-2);
        
          prefs.setCharPref("extensions.historyblock.stringpref", domainstring);
          prefs.setCharPref("extensions.historyblock.unblockedurl", "");
          return;
        }
      }
      // We always want this to happen no matter what
      prefs.setCharPref("extensions.historyblock.unblockedurl", "");
    }
  },
  
  /*
   * This is used to observe when the prefs get changed.
   */
  observe: function(subject, topic, state)
  {
    with (historyblock)
    {
      if(topic == "nsPref:changed" && state == "newblockedurl")
      {
        // Call the function to handle the additional hostname
        onHostnameAdd();
      }
      else if(topic == "nsPref:changed" && state == "unblockedurl")
      {
        // Call the function to handle the removed hostname
        onHostnameRemove();
      }
      else if(topic == "nsPref:changed" && state == "stringpref")
      {
        // For testing purposes only, really. If the user changes "about:config" for this pref, then
        // we need to update our loaded copy.
        domainstring = prefs.getCharPref("extensions.historyblock.stringpref");
      }
      else if(topic == "nsPref:changed" && state == "blockcookies")
      {
        // For testing purposes only, really. If the user changes "about:config" for this pref, then
        // we need to update our loaded copy.
        blockCookies = prefs.getBoolPref("extensions.historyblock.blockcookies");
      }
      else if(topic == "dl-done")
      {    
	    var oDownload = subject.QueryInterface(Components.interfaces.nsIDownload);
	    var domains = domainstring.split(",");
	    var url = oDownload.source.spec;
	    
	    var hashedTargetUrl = url;
	    hashedTargetUrl = hashedTargetUrl.match(hostRegexp);
	    var hostname = hashedTargetUrl[0].replace(slashesRegexp,"");

	    // Split on the '.'s
	    var subdomains = hostname.split(".");
	    if(subdomains.length > 2)
	    {
	      // Then we have some subdomain-ing going on.
	      if(prefs.getBoolPref("extensions.historyblock.blocksitewide"))
	      {
	        subdomains = subdomains[subdomains.length - 2] + "." + subdomains[subdomains.length - 1];
	      }
	    }
	    else
	    {
	      // Break it
	      subdomains = null;
	    }
	    
	    // Hash that hostname
	    hashedTargetUrl = hex_sha1(hostname);
	      
	    for(var blIndex=0; blIndex<domains.length; blIndex++)
	    {
	      if(hashedTargetUrl == domains[blIndex].replace(whiteSpace,""))
	      {
	        downloadManager.removeDownload(oDownload.id);
		    // Might as well clear out the history just in case...
		    removeHistory(hostname);
	        return;
	      }
	    }

	    if(subdomains == null) return;
	    // Hash that hostname
	    hashedTargetUrl = hex_sha1(subdomains);
	      
	    for(var blIndex=0; blIndex<domains.length; blIndex++)
	    {
	      if(hashedTargetUrl == domains[blIndex].replace(whiteSpace,""))
	      {
	        downloadManager.removeDownload(oDownload.id);
		    // Might as well clear out the history just in case...
		    removeHistory(hostname);
	        return;
	      }
	    }
      }
    }
  },
  
  /*
   * Helper function for getting all the cookies associated with a hostname.
   */
  getBlacklistedCookies: function(hostname)
  {
	with (historyblock)
	{
	  // Sentinel, don't bother if we aren't blocking cookies.
	  if(blockCookies) return new Array();
		
      var cookieEnum = cookieManager.enumerator;
      var blacklistCookies = new Array();
      
      while(cookieEnum.hasMoreElements())
      {
   	    var cookie = cookieEnum.getNext().QueryInterface(Components.interfaces.nsICookie2);
  	    var nextHost = cookie.host.replace("www\.","");
		    nextHost = nextHost.match(hostRegexp);
	    if(nextHost != null)
	    {
	      nextHost = nextHost[0];
    	  if(nextHost == hostname)
    	  {
    		blacklistCookies.push(cookie);
    	  }
	    }
      }
      return blacklistCookies;
	}
  },  
  
  /*
   * Helper function which adds back each nsICookie2 object
   * contained in blacklistCookies.
   */
  readdBlacklistCookies: function(blacklistCookies)
  {
	with (historyblock)
	{
	  // Sentinel, don't bother if we aren't blocking cookies.
	  if(blockCookies) return;
	  
      for(var i=0; i<blacklistCookies.length; i++)
      {
    	var cookie = blacklistCookies[i];
    	cookieManager.add(
    	  cookie.host,
    	  cookie.path,
    	  cookie.name,
    	  cookie.value,
    	  cookie.isSecure,
    	  cookie.isHttpOnly,
    	  cookie.isSession,
    	  cookie.expiry
        );
      }
	}
  },
  
  /*
   * A JavaScript implementation of the Secure Hash Algorithm, SHA-1, as defined
   * in FIPS PUB 180-1
   * Version 2.1a Copyright Paul Johnston 2000 - 2002.
   * Other contributors: Greg Holt, Andrew Kepert, Ydnar, Lostinet
   * Distributed under the BSD License
   * See http://pajhome.org.uk/crypt/md5 for details.
   */
   
  /*
   * These are the functions you'll usually want to call
   * They take string arguments and return either hex or base-64 encoded strings
   */
  hex_sha1: function(s)
  {
    with (historyblock)
    {
      return binb2hex(core_sha1(str2binb(s),s.length * chrsz));
    }
  },
  b64_sha1: function(s)
  {
    with (historyblock)
    {
      return binb2b64(core_sha1(str2binb(s),s.length * chrsz));
    }
  },
  str_sha1: function(s)
  {
    with (historyblock)
    {
      return binb2str(core_sha1(str2binb(s),s.length * chrsz));
    }
  },
  hex_hmac_sha1: function(key, data)
  {
    with (historyblock)
    {
      return binb2hex(core_hmac_sha1(key, data));
    }
  },
  b64_hmac_sha1: function(key, data)
  {
    with (historyblock)
    {
      return binb2b64(core_hmac_sha1(key, data));
    }
  },
  str_hmac_sha1: function(key, data)
  {
    with (historyblock)
    {
      return binb2str(core_hmac_sha1(key, data));
    }
  },

  /*
   * Calculate the SHA-1 of an array of big-endian words, and a bit length
   */
  core_sha1: function(x, len)
  {
    with (historyblock)
    {
      /* append padding */
      x[len >> 5] |= 0x80 << (24 - len % 32);
      x[((len + 64 >> 9) << 4) + 15] = len;
    
      var w = Array(80);
      var a =  1732584193;
      var b = -271733879;
      var c = -1732584194;
      var d =  271733878;
      var e = -1009589776;
    
      for(var i = 0; i < x.length; i += 16)
      {
        var olda = a;
        var oldb = b;
        var oldc = c;
        var oldd = d;
        var olde = e;
    
        for(var j = 0; j < 80; j++)
        {
          if(j < 16) w[j] = x[i + j];
          else w[j] = rol(w[j-3] ^ w[j-8] ^ w[j-14] ^ w[j-16], 1);
          var t = safe_add(safe_add(rol(a, 5), sha1_ft(j, b, c, d)),
                           safe_add(safe_add(e, w[j]), sha1_kt(j)));
          e = d;
          d = c;
          c = rol(b, 30);
          b = a;
          a = t;
        }
    
        a = safe_add(a, olda);
        b = safe_add(b, oldb);
        c = safe_add(c, oldc);
        d = safe_add(d, oldd);
        e = safe_add(e, olde);
      }
      return Array(a, b, c, d, e);
    }
  },

  core_sha1: function(x, len)
  {
    with (historyblock)
    {
      /* append padding */
      x[len >> 5] |= 0x80 << (24 - len % 32);
      x[((len + 64 >> 9) << 4) + 15] = len;
    
      var w = Array(80);
      var a =  1732584193;
      var b = -271733879;
      var c = -1732584194;
      var d =  271733878;
      var e = -1009589776;
    
      for(var i = 0; i < x.length; i += 16)
      {
        var olda = a;
        var oldb = b;
        var oldc = c;
        var oldd = d;
        var olde = e;
    
        for(var j = 0; j < 80; j++)
        {
          if(j < 16) w[j] = x[i + j];
          else w[j] = rol(w[j-3] ^ w[j-8] ^ w[j-14] ^ w[j-16], 1);
          var t = safe_add(safe_add(rol(a, 5), sha1_ft(j, b, c, d)),
                           safe_add(safe_add(e, w[j]), sha1_kt(j)));
          e = d;
          d = c;
          c = rol(b, 30);
          b = a;
          a = t;
        }
    
        a = safe_add(a, olda);
        b = safe_add(b, oldb);
        c = safe_add(c, oldc);
        d = safe_add(d, oldd);
        e = safe_add(e, olde);
      }
      return Array(a, b, c, d, e);
    }
  },
    
  /*
   * Perform the appropriate triplet combination function for the current
   * iteration
   */
  sha1_ft: function(t, b, c, d)
  {
    if(t < 20) return (b & c) | ((~b) & d);
    if(t < 40) return b ^ c ^ d;
    if(t < 60) return (b & c) | (b & d) | (c & d);
    return b ^ c ^ d;
  },
  
  /*
   * Perform the appropriate triplet combination function for the current
   * iteration
   */
  sha1_ft: function(t, b, c, d)
  {
    if(t < 20) return (b & c) | ((~b) & d);
    if(t < 40) return b ^ c ^ d;
    if(t < 60) return (b & c) | (b & d) | (c & d);
    return b ^ c ^ d;
  },
  
  /*
   * Determine the appropriate additive constant for the current iteration
   */
  sha1_kt: function(t)
  {
    return (t < 20) ?  1518500249 : (t < 40) ?  1859775393 :
           (t < 60) ? -1894007588 : -899497514;
  },

  /*
   * Calculate the HMAC-SHA1 of a key and some data
   */
  core_hmac_sha1: function(key, data)
  {
    with (historyblock)
    {
      var bkey = str2binb(key);
      if(bkey.length > 16) bkey = core_sha1(bkey, key.length * chrsz);
    
      var ipad = Array(16), opad = Array(16);
      for(var i = 0; i < 16; i++)
      {
        ipad[i] = bkey[i] ^ 0x36363636;
        opad[i] = bkey[i] ^ 0x5C5C5C5C;
      }
    
      var hash = core_sha1(ipad.concat(str2binb(data)), 512 + data.length * chrsz);
      return core_sha1(opad.concat(hash), 512 + 160);
    }
  },

  /*
   * Add integers, wrapping at 2^32. This uses 16-bit operations internally
   * to work around bugs in some JS interpreters.
   */
  safe_add: function(x, y)
  {
    var lsw = (x & 0xFFFF) + (y & 0xFFFF);
    var msw = (x >> 16) + (y >> 16) + (lsw >> 16);
    return (msw << 16) | (lsw & 0xFFFF);
  },

  /*
   * Bitwise rotate a 32-bit number to the left.
   */
  rol: function(num, cnt)
  {
    return (num << cnt) | (num >>> (32 - cnt));
  },

  /*
   * Convert an 8-bit or 16-bit string to an array of big-endian words
   * In 8-bit function, characters >255 have their hi-byte silently ignored.
   */
  str2binb: function(str)
  {
    with (historyblock)
    {
      var bin = Array();
      var mask = (1 << chrsz) - 1;
      for(var i = 0; i < str.length * chrsz; i += chrsz)
        bin[i>>5] |= (str.charCodeAt(i / chrsz) & mask) << (32 - chrsz - i%32);
      return bin;
    }
  },

  /*
   * Convert an array of big-endian words to a string
   */
  binb2str: function(bin)
  {
    with (historyblock)
    {
      var str = "";
      var mask = (1 << chrsz) - 1;
      for(var i = 0; i < bin.length * 32; i += chrsz)
        str += String.fromCharCode((bin[i>>5] >>> (32 - chrsz - i%32)) & mask);
      return str;
    }
  },

  /*
   * Convert an array of big-endian words to a hex string.
   */
  binb2hex: function(binarray)
  {
    with (historyblock)
    {
      var hex_tab = hexcase ? "0123456789ABCDEF" : "0123456789abcdef";
      var str = "";
      for(var i = 0; i < binarray.length * 4; i++)
      {
        str += hex_tab.charAt((binarray[i>>2] >> ((3 - i%4)*8+4)) & 0xF) +
               hex_tab.charAt((binarray[i>>2] >> ((3 - i%4)*8  )) & 0xF);
      }
      return str;
    }
  },

  /*
   * Convert an array of big-endian words to a base-64 string
   */
  binb2b64: function(binarray)
  {
    with (historyblock)
    {
      var tab = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
      var str = "";
      for(var i = 0; i < binarray.length * 4; i += 3)
      {
        var triplet = (((binarray[i   >> 2] >> 8 * (3 -  i   %4)) & 0xFF) << 16)
                    | (((binarray[i+1 >> 2] >> 8 * (3 - (i+1)%4)) & 0xFF) << 8 )
                    |  ((binarray[i+2 >> 2] >> 8 * (3 - (i+2)%4)) & 0xFF);
        for(var j = 0; j < 4; j++)
        {
          if(i * 8 + j * 6 > binarray.length * 32) str += b64pad;
          else str += tab.charAt((triplet >> 6*(3-j)) & 0x3F);
        }
      }
      return str;
    }
  }

};

// Gets called when the window loads, AKA when Firefox starts (or a new window is loaded).
window.addEventListener("load", function(){ historyblock.init(); }, false);
